#!/usr/bin/env python
"""
guano_edit.py - Manipulate GUANO metadata of individual files or in bulk.

Specify GUANO fields and values, followed by a list of files which the
changes should be applied to. The values of existing fields may be
used by specifying the field as `${Fieldname}`.

Be careful to properly escape values for your shell commandline, especially
if using value templates! Study the examples below.


Examples::

    # Add NABat grid cell to all recordings under the `foo` directory
    $> guano_edit.py  "NABat|Grid Cell ID: 45678"  ~/bat_calls/foo/

    # Append additional text to the end of the existing Note text
    $> guano_edit.py  'Note: ${Note} Recorded by Dave.'  EPFU_refcall.wav


TODO::
    * Ensure that we persist all RIFF chunks
    * Add support for adding GUANO metadata to "new" files
    * Support Anabat files
"""

from __future__ import print_function

import sys, os, os.path
from string import Template

import guano


DRY_RUN = False
MAKE_BACKUPS = True


class GuanoTemplate(Template):
    """
    String template with support for valid GUANO namespaced fields.
    """
    idpattern = r'[_a-z][_a-z0-9| ]*'  # added support for spaces and pipe char


def locate_files(rootdir):
    """Find files with GUANO metadata"""
    if os.path.isdir(rootdir):
        for root, dirnames, filenames in os.walk(rootdir):
            for filename in filenames:
                if filename.endswith('.wav') or filename.endswith('.WAV'):
                    try:
                        yield guano.GuanoFile(os.path.join(root, filename))
                    except ValueError as e:
                        pass
    elif os.path.isfile(rootdir):
        filename = rootdir
        try:
            yield guano.GuanoFile(filename)
        except ValueError as e:
            pass
    else:
        raise RuntimeError(rootdir)


def update(gfile, md, dry_run=False):
    """Update the GUANO metadata in a specified file"""
    print()
    print(gfile.filename)

    for k, v in md.items():
        gfile[k] = GuanoTemplate(v).substitute(gfile)

    print(gfile.to_string())
    if not dry_run:
        gfile.write(make_backup=MAKE_BACKUPS)


def main():
    """Commandline processing script"""
    md = {}      # new metadata values
    inputs = []  # files and folders we're operating on

    for arg in sys.argv[1:]:
        if ':' in arg:
            k, v = (x.strip() for x in arg.split(':', 1))
            md[k] = v
        else:
            inputs.append(arg)

    print(md)

    for input in inputs:
        for gfile in locate_files(input):
            update(gfile, md, dry_run=DRY_RUN)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Convert files with SonoBat-format metadata to use GUANO metadata instead.

usage::

    $> sb2guano.py WAVFILE...
"""

from __future__ import print_function

import sys
import os
import os.path
import mmap
import re
import wave
from contextlib import closing
from datetime import datetime
from pprint import pprint

from guano import GuanoFile


# regex for parsing Sonobat metadata
SB_MD_REGEX = re.compile(b'MMMMMMMMM(?P<sb_md>[\w\W]+)MMMMMMMMM')
SB_FREQ_REGEX = re.compile(r'\(#([\d]+)#\)')
SB_TE_REGEX = re.compile(r'<&([\d]*)&>')
SB_DFREQ_REGEX = re.compile(r'\[!([\w]+)!\]')

D500X_ATTRIBUTE_REGEX = re.compile(r'(?P<d500x>D500X V.+S/N=\d+)')
AR125_ATTRIBUTE_REGEX = re.compile(r'(?P<ar125>DEV=.+CMT=<.+>)')

# old SonoBat format e.g. TransectTestRun1-24Mar11-16,27,56-Myoluc.wav
SONOBAT_FILENAME1_REGEX = re.compile(r'(?P<date>[ 0123][0-9][A-Z][a-z][a-z][0-9][0-9]-[012][0-9],[0-6][0-9],[0-6][0-9])(-(?P<species>[A-Za-z]+))?')
SONOBAT_FILENAME1_TIMESTAMP_FMT = '%d%b%y-%H,%M,%S'

# new SonoBat format 4-digit year e.g. TransectTestRun1-20110324_162756-Myoluc.wav
SONOBAT_FILENAME2_REGEX = re.compile(r'(?P<date>\d{8}_\d{6})(-(?P<species>[A-Za-z]+))?')
SONOBAT_FILENAME2_TIMESTAMP_FMT = '%Y%m%d_%H%M%S'

# new new SonoBat format 2-digit year e.g. TransectTestRun1-20110324_162756-Myoluc.wav
SONOBAT_FILENAME3_REGEX = re.compile(r'(?P<date>\d{6}_\d{6})(-(?P<species>[A-Za-z]+))?')
SONOBAT_FILENAME3_TIMESTAMP_FMT = '%y%m%d_%H%M%S'

# AR125 raw
AR125_FILENAME_REGEX = re.compile(r'_(?P<date>D\d{8}T\d{6})m\d{3}(-(?P<species>[A-Za-z]+))?')
AR125_FILENAME_TIMESTAMP_FMT = 'D%Y%m%dT%H%M%S'

SB_FILENAME_FORMATS = [
    (SONOBAT_FILENAME1_REGEX, SONOBAT_FILENAME1_TIMESTAMP_FMT),
    (SONOBAT_FILENAME2_REGEX, SONOBAT_FILENAME2_TIMESTAMP_FMT),
    (SONOBAT_FILENAME3_REGEX, SONOBAT_FILENAME3_TIMESTAMP_FMT),
    (AR125_FILENAME_REGEX,    AR125_FILENAME_TIMESTAMP_FMT)
]


def _decode_text(text):
    """
    SonoBat uses the system locale for encoding text, so we have to guess what it might have been.
    Try Mac_Roman first if we're running on OS X, otherwise default to Windows 1252. Yuck.
    """
    encodings = ['windows-1252', 'latin-1']
    encodings.insert(0 if sys.platform == 'darwin' else 1, 'mac_roman')
    for encoding in encodings:
        try:
            return text.decode(encoding)
        except:
            pass
    raise ValueError('Unable to decode native SonoBat text!')


def _parse_sonobat_metadata(md):
    """Parse Sonobat-format metadata string as a dict"""
    sb_md = dict()
    sb_md['samplerate'] = int(re.search(SB_FREQ_REGEX, md).groups()[0])
    sb_md['te'] = int(re.search(SB_TE_REGEX, md).groups()[0])
    sb_md['dfreq'] = re.search(SB_DFREQ_REGEX, md).groups()[0]
    note = md.split('!]', 1)[1]

    # If this file was created with Sonobat D500X Attributer, parse out D500X metadata
    match = re.search(D500X_ATTRIBUTE_REGEX, note)
    if match and match.group('d500x').count(',') == 8:
        fw, f, pre, len, hp, a, ts, timestamp, sn = match.group('d500x').split(',')
        f, pre, len, hp, a, ts, sn = tuple(s.split('=',1)[1].strip() for s in (f, pre, len, hp, a, ts, sn))
        sb_md['d500x'] = dict(Firmware=fw, F=f, PRE=pre, LEN=len, HP=hp, A=a, TS=ts, Timestamp=timestamp, Serial=sn)

    # Binary Acoustic AR125 stuffs metadata into Sonobat note
    match = re.search(AR125_ATTRIBUTE_REGEX, note)
    if match:
        dev, dc, utc, ltb, cmt = match.group('ar125').split(',', 4)
        dev, dc, utc, ltb, cmt = tuple(s.split('=',1)[1].strip() for s in (dev, dc, utc, ltb, cmt))
        cmt = cmt.strip('<>')
        sb_md['ar125'] = dict(DEV=dev, DC=dc, UTC=utc, LTB=ltb, CMT=cmt)

    sb_md['note'] = note

    return sb_md


def extract_sonobat_metadata(fname):
    """Extract Sonobat-format metadata as a dict"""

    # parse the Sonobat metadata itself from file
    with open(fname, 'rb') as infile:
        with closing(mmap.mmap(infile.fileno(), 0, access=mmap.ACCESS_READ)) as mmfile:
            md_match = re.search(SB_MD_REGEX, mmfile)
            if not md_match:
                print('No Sonobat metadata found in file: ' + fname, file=sys.stderr)
                return None
            md = md_match.groups()[0]
            md = _decode_text(md)
            sb_md = _parse_sonobat_metadata(md)

    with closing(wave.open(fname)) as wavfile:
        duration_s = wavfile.getnframes() / float(wavfile.getframerate())
        sb_md['length'] = round(duration_s / sb_md['te'], 2)

    # try to extract info from the filename
    for regex, timestamp_fmt in SB_FILENAME_FORMATS:
        match = regex.search(fname)
        if match:
            sb_md['timestamp'] = datetime.strptime(match.group('date'), timestamp_fmt)
            sb_md['species'] = match.group('species')

    return sb_md


def sonobat2guano(fname):
    """Convert a file with Sonobat metadata to GUANO metadata"""
    print('\n', fname)
    sb_md = extract_sonobat_metadata(fname)
    if not sb_md:
        print('Skipping non-Sonobat file: ' + fname, file=sys.stderr)
        return False
    pprint(sb_md)

    gfile = GuanoFile(fname)
    gfile['GUANO|Version'] = 1.0
    if 'timestamp' in sb_md:
        gfile['Timestamp'] = sb_md['timestamp']
    if sb_md.get('te', 1) != 1:
        gfile['TE'] = sb_md['te']
    gfile['Length'] = sb_md['length']
    gfile['Note'] = sb_md['note'].strip().replace('\r\n', '\\n').replace('\n', '\\n')
    if sb_md.get('species', None):
        gfile['Species Auto ID'] = sb_md['species']

    if 'd500x' in sb_md:
        for k, v in sb_md['d500x'].items():
            gfile['PET', k] = v

    if 'ar125' in sb_md:
        for k, v in sb_md['ar125'].items():
            gfile['BAT', k] = v

    print(gfile.to_string())

    gfile.write()


if __name__ == '__main__':
    from glob import glob

    if len(sys.argv) < 2:
        print('usage: %s FILE...' % os.path.basename(sys.argv[0]), file=sys.stderr)
        sys.exit(2)

    if os.name == 'nt' and '*' in sys.argv[1]:
        fnames = glob(sys.argv[1])
    else:
        fnames = sys.argv[1:]

    for fname in fnames:
        sonobat2guano(fname)



#!/usr/bin/env python
"""
"Disperse" files by moving them into folders according to their species label.

The `Species Manual ID` field will be preferred over `Species Auto ID`.

usage::

    $> disperse.py [--copy] ROOTDIR
"""

# TODO: distinguish between Manual / Auto ID; un-disperse; recursive

from __future__ import print_function

import os
import os.path
from glob import glob
import shutil

import guano


def get_species(fname):
    """Get the species label from a GUANO file, or `None`. Prefer `Manual ID` over 'Auto ID'."""
    try:
        f = guano.GuanoFile(fname)
    except ValueError:
        return None

    species = f.get('Species Manual ID', None)
    return species if species else f.get('Species Auto ID', None)


def disperse(rootdir, copy=False, destination_root=None):
    """
    Disperse GUANO .wav files into folders by their species label.

    :param str rootdir:  the root directory where we search for GUANO files
    :param bool copy:    whether we should *copy* or *move* (default) files
    :param str destination_root:  optional destination directory where files are output
    """
    for fname in glob(os.path.join(rootdir, '*.wav')):
        species = get_species(fname)
        if not species:
            print('Skipping file without species %s .' % fname)
            continue
        destination = os.path.join(destination_root or rootdir, species)
        if not os.path.isdir(destination):
            print('Creating directory %s ...' % destination)
            os.mkdir(destination)
        new_fname = os.path.join(destination, os.path.basename(fname))
        print('%sing %s -> %s ...' % ('copy' if copy else 'mov', fname, new_fname))
        if copy:
            shutil.copy2(fname, new_fname)
        else:
            os.rename(fname, new_fname)


def main():
    """Commandline interface"""
    import argparse
    parser = argparse.ArgumentParser(description='Disperse files to folders by their species field')
    parser.add_argument('-c', '--copy', action='store_true', help='Copy files rather than moving them')
    parser.add_argument('rootdir')
    args = parser.parse_args()
    disperse(args.rootdir, copy=args.copy)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Print the GUANO metadata found in a file or files.

usage::

    $> guano_dump.py [--strict] WAVFILE...
"""

from __future__ import print_function

import sys
import os
import os.path

from guano import GuanoFile


def dump(fname, strict=False):
    print()
    print(fname)
    gfile = GuanoFile(fname, strict=strict)
    print(gfile.to_string())


if __name__ == '__main__':
    from glob import glob
    import logging
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s\t%(levelname)s\t%(message)s')

    if len(sys.argv) < 2:
        print('usage: %s [--strict] FILE...' % os.path.basename(sys.argv[0]), file=sys.stderr)
        sys.exit(2)

    if os.name == 'nt' and '*' in sys.argv[1]:
        fnames = glob(sys.argv[1])
    else:
        fnames = sys.argv[1:]

    strict = False
    if '--strict' in fnames:
        fnames.remove('--strict')
        strict = True

    for fname in fnames:
        if os.path.isdir(fname):
            for subfname in glob(os.path.join(fname, '*.[Ww][Aa][Vv]')):
                dump(subfname, strict=strict)
        else:
            dump(fname, strict=strict)



#!/usr/bin/env python
"""
Convert files with raw D500X metadata to use GUANO metadata instead.

usage::

    $> d500x2guano.py WAVFILE...
"""

from __future__ import print_function

import sys
import os
import os.path
import mmap
import wave
import stat
import struct
from contextlib import closing
from datetime import datetime
from pprint import pprint

from guano import GuanoFile


D500X_DATA_SKIP_BYTES = 0x3D4


def dms2decimal(dms_str):
    """Convert D500X Degrees-Minuts-Seconds to Decimal Degrees"""
    d, m, s, direction = dms_str.split()
    sign = -1 if direction in ('S', 'W') else 1
    return sign * (int(d) + float(m) / 60 + float(s) / 3600)


def unlock(fname):
    """Enable filesystem modification of a locked file"""
    if os.name == 'nt':
        os.chmod(fname, stat.S_IWRITE)
    elif os.name == 'posix' and hasattr(os, 'chflags'):
        os.chflags(fname, os.stat(fname).st_flags & ~stat.UF_IMMUTABLE)


def extract_d500x_metadata(fname):
    """Extract raw D500X metadata as a dict, or None if file has none"""
    md = {}
    with open(fname, 'rb') as infile:
        with closing(mmap.mmap(infile.fileno(), 0, access=mmap.ACCESS_READ)) as mmfile:
            if mmfile[0xF0:0xF0+5] != b'D500X':
                print('No D500X metadata found in file: ' + fname, file=sys.stderr)
                return None

            md['Samplerate'] = struct.unpack_from('< i', mmfile, 0x18)[0]
            md['File Name'] = mmfile[0xD0:0xD0+10].decode('latin-1')
            md['File Time'] = mmfile[0xE0:0xE0+15].decode('latin-1')
            md['FW Version'] = mmfile[0xF0:0xF0+32].strip(b'\0 ').decode('latin-1')
            profile_settings_1 = mmfile[0x120:0x120+20].strip(b'\0 ').decode('latin-1')
            profile_settings_2 = mmfile[0x138:0x138+16].strip(b'\0 ').decode('latin-1')
            for tok in (profile_settings_1 + ' ' + profile_settings_2).split():
                k, v = tok.split('=', 1)
                md['Profile ' + k] = v
            # TODO:  0x150 - 0x157 ?
            md['Profile Name'] = mmfile[0x158:0x158+8].strip(b'\0\xFF ').decode('latin-1')

            # block from 0x200 - 0x400 is a big '\r\n' delimited string. 2.0+ firmware only
            extra_md_block = mmfile[0x200:0x400].strip().decode('latin-1')
            if extra_md_block:
                for line in extra_md_block.splitlines():
                    if not line.strip('\0 '):
                        continue
                    k, v = line.split(':', 1)
                    md[k] = v.strip('\0 ')

            md['File Time'] = datetime.strptime(md['File Time'], '%y%m%d %H:%M:%S')

    with closing(wave.open(fname)) as wavfile:
        frame_count = wavfile.getnframes() - (D500X_DATA_SKIP_BYTES / wavfile.getsampwidth())
        duration_s = frame_count / float(wavfile.getframerate())
        md['Length'] = round(duration_s, 2)

    return md


def d500x2guano(fname):
    """Convert a file with raw D500X metadata to use GUANO metadata instead"""
    print('\n', fname)
    md = extract_d500x_metadata(fname)
    if not md:
        print('Skipping non-D500X file: ' + fname, file=sys.stderr)
        return False
    #pprint(md)

    gfile = GuanoFile(fname)
    gfile['GUANO|Version'] = 1.0

    gfile['Make'] = 'Pettersson'
    gfile['Model'] = 'D500X'
    gfile['Timestamp'] = md.pop('File Time')
    gfile['Original Filename'] = md.pop('File Name')
    gfile['Samplerate'] = md.pop('Samplerate')
    gfile['Length'] = md.pop('Length')

    if md.get('Profile HP', None) == 'Y':
        gfile['Filter HP'] = 20

    lat, lon = md.pop('LAT', None), md.pop('LON', None)
    if lat and lon:
        gfile['Loc Position'] = dms2decimal(lat), dms2decimal(lon)

    for k, v in md.items():
        gfile['PET', k] = v

    print(gfile.to_string())

    # throw out the Pettersson metadata bytes from 'data' chunk
    gfile.wav_data = gfile.wav_data[D500X_DATA_SKIP_BYTES:]

    unlock(fname)  # D500X "locks" files as unwriteable, we must unlock before we can modify
    gfile.write()


if __name__ == '__main__':
    from glob import glob

    if len(sys.argv) < 2:
        print('usage: %s FILE...' % os.path.basename(sys.argv[0]), file=sys.stderr)
        sys.exit(2)

    if os.name == 'nt' and '*' in sys.argv[1]:
        fnames = glob(sys.argv[1])
    else:
        fnames = sys.argv[1:]

    for fname in fnames:
        d500x2guano(fname)



"""
This is the Python reference implementation for reading and writing GUANO metadata.

GUANO is the "Grand Unified Acoustic Notation Ontology", an extensible metadata format
for representing bat acoustics data.

Import this Python module as::

    import guano

This module utilizes the Python :mod:`logging` framework for issuing warnings and debug messages.
Application code may wish to enable logging with the :func:`logging.basicConfig` function.

"""

import os
import sys
import wave
import struct
import os.path
import shutil
from datetime import datetime, tzinfo, timedelta
from contextlib import closing
from tempfile import NamedTemporaryFile
from collections import OrderedDict, namedtuple
from base64 import standard_b64encode as base64encode
from base64 import standard_b64decode as base64decode

import logging
log = logging.Logger(__name__)

if sys.version_info[0] > 2:
    unicode = str
    basestring = str


__version__ = '1.0.14'

__all__ = 'GuanoFile',


WHITESPACE = ' \t\n\x0b\x0c\r\0'

wavparams = namedtuple('wavparams', 'nchannels, sampwidth, framerate, nframes, comptype, compname')


_ZERO = timedelta(0)

class tzutc(tzinfo):
    """UTC timezone"""

    def utcoffset(self, dt):
        return _ZERO

    def tzname(self, dt):
        return 'UTC'

    def dst(self, dt):
        return _ZERO

    def __repr__(self):
        return 'UTC'

utc = tzutc()

class tzoffset(tzinfo):
    """
    Fixed-offset concrete timezone implementation.
    `offset` should be numeric hours or ISO format string like '-07:00'.
    """

    def __init__(self, offset):
        if isinstance(offset, basestring):
            # offset as ISO string '-07:00', '-0700', or '-07' format
            if len(offset) < 4:
                vals = offset, '00'              # eg '-07'
            elif ':' in offset:
                vals = offset.split(':')         # '-07:00'
            else:
                vals = offset[:-2], offset[-2:]  # '-0700'
            if vals[0].startswith('-'):
                offset = int(vals[0]) - int(vals[1])/60.0
            else:
                offset = int(vals[0]) + int(vals[1])/60.0
        self._offset_hours = offset
        self._offset = timedelta(hours=offset)

    def utcoffset(self, dt):
        return self._offset

    def dst(self, dt):
        return _ZERO

    def tzname(self, dt):
        return 'UTC'+str(self._offset_hours)

    def __repr__(self):
        return self.tzname(None)


def parse_timestamp(s):
    """
    Parse a string in supported subset of ISO 8601 / RFC 3331 format to :class:`datetime.datetime`.
    The timestamp will be timezone-aware of a TZ is specified, or timezone-naive if in "local" fmt.

    :rtype: datetime or None
    """
    # Python's standard library does an awful job of parsing ISO timestamps, so we do it manually

    if s is None or not s.strip():
        return None

    timestamp, tz = None, None

    s = s.replace(' ', 'T', 1)  # support using space rather than 'T' as date/time delimiter

    if s[-1] == 'Z':  # UTC "zulu" time
        tz = utc
        s = s[:-1]
    elif '+' in s or s.count('-') == 3:  # UTC offset provided
        i = s.index('+') if '+' in s else s.rfind('-')
        s, offset = s[:i], s[i:]
        tz = tzoffset(offset)

    if len(s) > 22:  # milliseconds included
        timestamp = datetime.strptime(s, '%Y-%m-%dT%H:%M:%S.%f')
    else:
        timestamp = datetime.strptime(s, '%Y-%m-%dT%H:%M:%S')

    return timestamp.replace(tzinfo=tz) if tz else timestamp


_chunkid = struct.Struct('> 4s')
_chunksz = struct.Struct('< L')


class GuanoFile(object):
    """
    An abstraction of a .WAV file with GUANO metadata.

    A `GuanoFile` object behaves like a normal Python :class:`dict`, where keys can either be
    well-known metadata keys, namespaced keys, or a tuple of (namespace, key).

    Well-known keys will have their values coerced into the correct data type. The parser may be
    configured to coerce new namespaced keys with the :func:`register()` function.

    Example usage::

        gfile = GuanoFile('myfile.wav')
        print gfile['GUANO|Version']
        >>> '1.0'
        gfile['Species Manual ID'] = 'Mylu'
        gfile['Note'] = 'I love GUANO!'
        gfile.write()

    Though reading, writing, and editing .WAV files is the target usage, this class may also be
    used independent from the .WAV file format. GUANO metadata can be written into an
    Anabat-format file or to a sidecar file, for example, by populating a `GuanoFile` object and
    then using the :func:`serialize()` method to produce correctly formatted UTF-8 encoded metadata.

    :ivar str filename:  path to the file which this object represents, or `None` if a "new" file
    :ivar bool strict_mode:  whether the GUANO parser is configured for strict or lenient parsing
    :ivar bytes wav_data:  the `data` subchunk of a .WAV file consisting of its actual audio data,
                           lazily-loaded and cached for performance
    :ivar wavparams wav_params:  namedtuple of .WAV parameters (nchannels, sampwidth, framerate, nframes, comptype, compname)
    """

    _coersion_rules = {
        'Filter HP': float, 'Length': float, 'Loc Elevation': float,
        'Loc Accuracy': int, 'Samplerate': int,
        'TE': lambda value: int(value) if value else 1,
        'Loc Position': lambda value: tuple(float(v) for v in value.split()),
        'Timestamp': parse_timestamp,
        'Note': lambda value: value.replace('\\n', '\n'),
    }
    _serialization_rules = {
        'Loc Position': lambda value: '%f %f' % value,
        'Timestamp': lambda value: value.isoformat() if value else '',
        'Length': lambda value: '%.2f' % value,
        'Note': lambda value: value.replace('\n', '\\n')
    }

    def __init__(self, filename=None, strict=True):
        """
        Create a GuanoFile instance which represents a single file's GUANO metadata.
        If the file already contains GUANO metadata, it will be parsed immediately. If not, then
        this object will be initialized as "new" metadata.

        :param filename:  path to an existing .WAV file with GUANO metadata; if the path does not
                          exist or is `None` then this instance represents a "new" file
        :type filename:  str or None
        :param bool strict:  whether the parser should be strict and raise exceptions when
                             encountering bad metadata values, or whether it should be as lenient
                             as possible (default: True); if in lenient mode, bad values will
                             remain in their UTF-8 string form as found persisted in the file
        :raises ValueError:  if the specified file doesn't represent a valid .WAV or if its
                             existing GUANO metadata is broken
        """
        self.filename = filename
        self.strict_mode = strict

        self.wav_params = None
        self._md = OrderedDict()  # metadata storage - map of maps:  namespace->key->val

        self._wav_data = None  # lazily-loaded and cached
        self._wav_data_offset = 0
        self._wav_data_size = 0

        if filename is not None and os.path.isfile(filename):
            self._load()

    def _coerce(self, key, value):
        """Coerce a value from its Unicode representation to a specific data type"""
        if key in self._coersion_rules:
            try:
                return self._coersion_rules[key](value)
            except (ValueError, TypeError) as e:
                if self.strict_mode:
                    raise
                else:
                    log.warning('Failed coercing "%s": %s', key, e)
        return value  # default should already be a Unicode string

    def _serialize(self, key, value):
        """Serialize a value from its real representation to GUANO Unicode representation"""
        serialize = self._serialization_rules.get(key, unicode)
        try:
            return serialize(value)
        except (ValueError, TypeError) as e:
            if self.strict_mode:
                raise
            else:
                log.warning('Failed serializing "%s": %s', key, e)

    def _load(self):
        """Load the contents of our underlying .WAV file"""
        fsize = os.path.getsize(self.filename)
        if fsize < 8:
            raise ValueError('File too small to contain valid RIFF "WAVE" header (size %d bytes)' % fsize)

        with open(self.filename, 'rb') as f:

            f.seek(0x08)
            chunk = _chunkid.unpack(f.read(4))[0]
            if chunk != b'WAVE':
                raise ValueError('Expected RIFF chunk "WAVE" at 0x08, but found "%s"' % repr(chunk))

            try:
                f.seek(0)
                self.wav_params = wavparams(*wave.open(f).getparams())
            except RuntimeError as e:
                return ValueError(e)  # Python's chunk.py throws this inappropriate exception

            # iterate through the file until we find our 'guan' subchunk
            metadata_buf = None
            f.seek(0x0c)
            while f.tell() < fsize - 1:
                try:
                    chunkid = _chunkid.unpack(f.read(4))[0]
                    size = _chunksz.unpack(f.read(4))[0]
                except struct.error as e:
                    raise ValueError(e)

                if chunkid == b'guan':
                    metadata_buf = f.read(size)
                elif chunkid == b'data':
                    self._wav_data_offset = f.tell()
                    self._wav_data_size = size
                    f.seek(size, 1)
                else:
                    f.seek(size, 1)

                if size % 2:
                    f.read(1)  # align to 16-bit boundary

            if not self._wav_data_offset:
                raise ValueError('No DATA sub-chunk found in .WAV file')

            if metadata_buf:
                self._parse(metadata_buf)

    def _parse(self, metadata_str):
        """Parse metadata and populate our internal mappings"""
        if not isinstance(metadata_str, unicode):
            try:
               metadata_str = metadata_str.decode('utf-8')
            except UnicodeDecodeError as e:
                log.warning('GUANO metadata is not UTF-8 encoded! Attempting to coerce. %s', self.filename)
                metadata_str = metadata_str.decode('latin-1')

        for line in metadata_str.split('\n'):
            line = line.strip(WHITESPACE)
            if not line:
                continue
            full_key, val = line.split(':', 1)
            namespace, key = full_key.split('|', 1) if '|' in full_key else ('', full_key)
            namespace, key, full_key, val = namespace.strip(), key.strip(), full_key.strip(), val.strip()
            if not key or not val:
                continue
            if namespace not in self._md:
                self._md[namespace] = OrderedDict()
            self._md[namespace][key] = self._coerce(full_key, val)
        return self

    @classmethod
    def from_string(cls, metadata_str, *args, **kwargs):
        """
        Create a :class:`GuanoFile` instance from a GUANO metadata string

        :param metadata_str:  a string (or string-like buffer) of GUANO metadata
        :param bool strict:  whether the parser should be strict and raise exceptions when
                             encountering bad metadata values, or whether it should be as lenient
                             as possible (default: True); if in lenient mode, bad values will
                             remain in their UTF-8 string form as found persisted in the file
        :rtype:  GuanoFile
        """
        return GuanoFile(*args, **kwargs)._parse(metadata_str)

    @classmethod
    def register(cls, namespace, keys, coerce_function, serialize_function=str):
        """
        Configure the GUANO parser to recognize new namespaced keys.

        :param namespace:  vendor namespace which the keys belong to
        :param keys:  a key or sequence of keys under the specified vendor namespace
        :param coerce_function:  a function for coercing the UTF-8 value to any desired data type
        :type coerce_function:  callable
        :param serialize_function:  an optional function for serializing the value to UTF-8 string
        :type serialize_function:  callable
        """
        if isinstance(keys, basestring):
            keys = [keys]
        for k in keys:
            full_key = namespace+'|'+k if namespace else k
            cls._coersion_rules[full_key] = coerce_function
            cls._serialization_rules[full_key] = serialize_function

    def _split_key(self, item):
        if isinstance(item, tuple):
            namespace, key = item[0], item[1]
        elif '|' in item:
            namespace, key = item.split('|', 1)
        else:
            namespace, key = '', item
        return namespace, key

    def __getitem__(self, item):
        namespace, key = self._split_key(item)
        return self._md[namespace][key]

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default

    def __setitem__(self, key, value):
        if not self._md:
            self._md['GUANO'] = {}
            self._md['GUANO']['Version'] = '1.0'

        namespace, key = self._split_key(key)
        if namespace not in self._md:
            self._md[namespace] = {}
        self._md[namespace][key] = value

    def __contains__(self, item):
        namespace, key = self._split_key(item)
        return namespace in self._md and key in self._md[namespace]

    def __delitem__(self, key):
        namespace, key = self._split_key(key)
        del self._md[namespace][key]
        if not self._md[namespace]:
            del self._md[namespace]

    def __bool__(self):
        return bool(self._md)
    __nonzero__ = __bool__  # py2

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.filename)

    def get_namespaces(self):
        """
        Get list of all namespaces represented by this metadata.
        This includes the 'GUANO' namespace, and the '' (empty string) namespace for well-known fields.
        """
        return self._md.keys()

    def items(self, namespace=None):
        """Iterate over (key, value) for entire metadata or for specified namespace of fields"""
        if namespace is not None:
            for k, v in self._md[namespace].items():
                yield k, v
        else:
            for namespace, data in self._md.items():
                for k, v in data.items():
                    k = '%s|%s' % (namespace, k) if namespace else k
                    yield k, v

    def items_namespaced(self):
        """Iterate over (namespace, key, value) for entire metadata"""
        for namespace, data in self._md.items():
            for k, v in data.items():
                yield namespace, k, v

    def well_known_items(self):
        """Iterate over (key, value) for all the well-known (defined) fields"""
        return self.items('')

    def to_string(self):
        """Represent the GUANO metadata as a Unicode string"""
        lines = []
        for namespace, data in self._md.items():
            for k, v in data.items():
                k = u'%s|%s' % (namespace, k) if namespace else k
                v = self._serialize(k, v)
                lines.append(u'%s: %s' % (k, v))
        return u'\n'.join(lines)

    def serialize(self, pad='\n'):
        """Serialize the GUANO metadata as UTF-8 encoded bytes"""
        md_bytes = bytearray(self.to_string(), 'utf-8')
        if pad is not None and len(md_bytes) % 2:
            # pad for alignment on even word boundary
            md_bytes.append(ord(pad))
        return md_bytes

    @property
    def wav_data(self):
        """Actual audio data from the wav `data` chunk. Lazily loaded and cached."""
        if not self._wav_data_size:
            raise ValueError()
        if not self._wav_data:
            with open(self.filename, 'rb') as f:
                f.seek(self._wav_data_offset)
                self._wav_data = f.read(self._wav_data_size)
        return self._wav_data

    @wav_data.setter
    def wav_data(self, data):
        self._wav_data_size = len(data)
        self._wav_data = data

    def write(self, make_backup=True):
        """
        Write the GUANO .WAV file to disk.

        :param bool make_backup:  create a backup file copy before writing changes or not (default: True);
                                  backups will be saved to a folder named `GUANO_BACKUP`
        :raises ValueError:  if this `GuanoFile` doesn't represent a valid .WAV by having
            appropriate values for `self.wav_params` (see :meth:`wave.Wave_write.setparams()`)
            and `self.wav_data` (see :meth:`wave.Wave_write.writeframes()`)
        """
        # FIXME: optionally write other unknown subchunks for redundant metadata formats

        if not self.filename:
            raise ValueError('Cannot write .WAV file without a self.filename!')
        if not self.wav_params:
            raise ValueError('Cannot write .WAV file without appropriate self.wav_params (see `wavfile.setparams()`)')
        if not self.wav_data:
            raise ValueError('Cannot write .WAV file without appropriate self.wav_data (see `wavfile.writeframes()`)')

        # prepare our metadata for a byte-wise representation
        md_bytes = self.serialize()

        # create tempfile and write our vanilla .WAV ('data' sub-chunk only)
        tempfile = NamedTemporaryFile(mode='w+b', prefix='guano_temp-', suffix='.wav', delete=False)
        if os.path.isfile(self.filename):
            shutil.copystat(self.filename, tempfile.name)

        with closing(wave.Wave_write(tempfile)) as wavfile:
            wavfile.setparams(self.wav_params)
            wavfile.writeframes(self.wav_data)

        # add the 'guan' sub-chunk after the 'data' sub-chunk
        tempfile.write(_chunkid.pack(b'guan'))
        tempfile.write(_chunksz.pack(len(md_bytes)))
        tempfile.write(md_bytes)

        # fix the RIFF file length
        total_size = tempfile.tell()
        tempfile.seek(0x04)
        tempfile.write(_chunksz.pack(total_size - 8))
        tempfile.close()

        # verify it by re-parsing the new version
        GuanoFile(tempfile.name)

        # finally overwrite the original with our new version (and optionally back up first)
        if make_backup and os.path.exists(self.filename):
            backup_dir = os.path.join(os.path.dirname(self.filename), 'GUANO_BACKUP')
            backup_file = os.path.join(backup_dir, os.path.basename(self.filename))
            if not os.path.isdir(backup_dir):
                log.debug('Creating backup dir: %s', backup_dir)
                os.mkdir(backup_dir)
            if os.path.exists(backup_file):
                os.remove(backup_file)
            shutil.move(self.filename, backup_file)
        shutil.move(tempfile.name, self.filename)


# This ugly hack prevents a warning if application-level code doesn't configure logging
if sys.version_info[0] > 2:
    NullHandler = logging.NullHandler
else:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass
log.addHandler(NullHandler())



import io
import sys
from setuptools import setup
from glob import glob

from guano import __version__


setup(
    name='guano',
    version=__version__,
    description='GUANO, the "Grand Unified" bat acoustics metadata format',
    long_description=io.open('README.rst', encoding='utf-8').read(),
    url='https://github.com/riggsd/guano-py',
    license='MIT',
    author='David A. Riggs',
    author_email='driggs@myotisoft.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
    keywords='bats acoustics metadata guano',
    py_modules=['guano'],
    scripts=glob('bin/*.py'),
)



#!/usr/bin/python
"""
Convert Wildlife Acoustics WAMD metadata files to use GUANO metadata instead.

usage::

    $> wamd2guano.py WAVFILE...
"""

from __future__ import print_function

import os
import os.path
import sys
import chunk
import struct
from datetime import datetime
from pprint import pprint

from guano import GuanoFile, tzoffset


# binary WAMD field identifiers
WAMD_IDS = {
    0x00: 'version',
    0x01: 'model',
    0x02: 'serial',
    0x03: 'firmware',
    0x04: 'prefix',
    0x05: 'timestamp',
    0x06: 'gpsfirst',
    0x07: 'gpstrack',
    0x08: 'software',
    0x09: 'license',
    0x0A: 'notes',
    0x0B: 'auto_id',
    0x0C: 'manual_id',
    0x0D: 'voicenotes',
    0x0E: 'auto_id_stats',
    0x0F: 'time_expansion',
    0x10: 'program',
    0x11: 'runstate',
    0x12: 'microphone',
    0x13: 'sensitivity',
}

# fields that we exclude from our in-memory representation
WAMD_DROP_IDS = (
    0x0D,    # voice note embedded .WAV
    0x10,    # program binary
    0x11,    # runstate giant binary blob
    0xFFFF,  # used for 16-bit alignment
)

# rules to coerce values from binary string to native types (default is `str`)
WAMD_COERCE = {
    'version': lambda x: struct.unpack('<H', x)[0],
    'timestamp': lambda x: _parse_wamd_timestamp(x),
    'gpsfirst': lambda x: _parse_wamd_gps(x),
    'time_expansion': lambda x: struct.unpack('<H', x)[0],
}


def _parse_text(value):
    """Default coercion function which assumes text is UTF-8 encoded"""
    return value.decode('utf-8')


def _parse_wamd_timestamp(timestamp):
    """WAMD timestamps are one of these known formats:
    2014-04-02 22:59:14-05:00
    2014-04-02 22:59:14.000
    2014-04-02 22:59:14
    Produces a `datetime.datetime`.
    """
    if isinstance(timestamp, bytes):
        timestamp = timestamp.decode('utf-8')
    if len(timestamp) == 25:
        dt, offset = timestamp[:-6], timestamp[19:]
        tz = tzoffset(offset)
        return datetime.strptime(dt, '%Y-%m-%d %H:%M:%S').replace(tzinfo=tz)
    elif len(timestamp) == 23:
        return datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S.%f')
    elif len(timestamp) == 19:
        return datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S')
    else:
        return None


def _parse_wamd_gps(gpsfirst):
    """WAMD "GPS First" waypoints are in one of these two formats:
    SM3, SM4, (the correct format):
        WGS..., LAT, N|S, LON, E|W [, alt...]
    EMTouch:
        WGS..., [-]LAT, [-]LON[,alt...]
    Produces (lat, lon, altitude) float tuple.
    """
    if not gpsfirst:
        return None
    if isinstance(gpsfirst, bytes):
        gpsfirst = gpsfirst.decode('utf-8')
    vals = tuple(val.strip() for val in gpsfirst.split(','))
    datum, vals = vals[0], vals[1:]
    if vals[1] in ('N', 'S'):
        # Standard format
        lat, lon = float(vals[0]), float(vals[2])
        if vals[1] == 'S':
            lat *= -1
        if vals[3] == 'W':
            lon *= -1
        alt = int(round(float(vals[4]))) if len(vals) > 4 else None
    else:
        # EMTouch format
        lat, lon = float(vals[0]), float(vals[1])
        alt = int(round(float(vals[2]))) if len(vals) > 2 else None
    return lat, lon, alt


def wamd(fname):
    """Extract WAMD metadata from a .WAV file as a dict"""
    with open(fname, 'rb') as f:
        ch = chunk.Chunk(f, bigendian=False)
        if ch.getname() != b'RIFF':
            raise Exception('%s is not a RIFF file!' % fname)
        if ch.read(4) != b'WAVE':
            raise Exception('%s is not a WAVE file!' % fname)

        wamd_chunk = None
        while True:
            try:
                subch = chunk.Chunk(ch, bigendian=False)
            except EOFError:
                break
            if subch.getname() == b'wamd':
                wamd_chunk = subch
                break
            else:
                subch.skip()
        if not wamd_chunk:
            raise Exception('"wamd" WAV chunk not found in file %s' % fname)

        metadata = {}
        offset = 0
        size = wamd_chunk.getsize()
        buf = wamd_chunk.read(size)
        while offset < size:
            id = struct.unpack_from('< H', buf, offset)[0]
            len = struct.unpack_from('< I', buf, offset+2)[0]
            val = struct.unpack_from('< %ds' % len, buf, offset+6)[0]
            if id not in WAMD_DROP_IDS:
                name = WAMD_IDS.get(id, id)
                val = WAMD_COERCE.get(name, _parse_text)(val)
                metadata[name] = val
            offset += 6 + len
        return metadata


def wamd2guano(fname):
    """Convert a Wildlife Acoustics WAMD metadata file to GUANO metadata format"""
    wamd_md = wamd(fname)
    pprint(wamd_md)

    gfile = GuanoFile(fname)
    gfile['GUANO|Version'] = 1.0

    gfile['Timestamp'] = wamd_md.pop('timestamp')
    gfile['Note'] = wamd_md.pop('notes', '')

    gfile['Make'] = 'Wildlife Acoustics'
    gfile['Model'] = wamd_md.pop('model', '')
    gfile['Firmware Version'] = wamd_md.pop('firmware', '')

    gfile['Species Auto ID'] = wamd_md.pop('auto_id', '')
    gfile['Species Manual ID'] = wamd_md.pop('manual_id', '')

    gfile['TE'] = wamd_md.pop('time_expansion', 1)
    gfile['Samplerate'] = gfile.wav_params.framerate * gfile['TE']
    gfile['Length'] = gfile.wav_params.nframes / float(gfile.wav_params.framerate) * gfile['TE']

    if 'gpsfirst' in wamd_md:
        lat, lon, alt = wamd_md.pop('gpsfirst')
        gfile['Loc Position'] = lat, lon
        gfile['Loc Elevation'] = alt

    for k, v in wamd_md.items():
        gfile['WA', k] = v

    print(gfile.to_string())

    gfile.write()


def main():
    from glob import glob

    if len(sys.argv) < 2:
        print('usage: %s FILE...' % os.path.basename(sys.argv[0]), file=sys.stderr)
        sys.exit(2)

    if os.name == 'nt' and '*' in sys.argv[1]:
        fnames = glob(sys.argv[1])
    else:
        fnames = sys.argv[1:]

    for fname in fnames:
        print(fname)
        try:
            wamd2guano(fname)
        except Exception as e:
            import traceback
            traceback.print_exc()
            #print(e, file=sys.stderr)
        print()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Convert files from the Elekon BatLogger to use GUANO metadata instead.

usage::

    $> batlogger2guano.py WAVFILE...
"""

from __future__ import print_function

import sys, os, os.path
from datetime import datetime
from xml.etree import ElementTree

from guano import GuanoFile


def get(xml, path, coerce=None, default=None):
    """Extract a value from an ElementTree node"""
    node = xml.find(path)
    if node is None:
        return default
    if coerce is None:
        return node.text
    else:
        return coerce(node.text)


def batlogger2guano(fname):
    """Convert an Elekon BatLogger .WAV with sidecar .XML to GUANO metadata"""
    xmlfname = os.path.splitext(fname)[0] + '.xml'
    if not os.path.exists(xmlfname):
        raise ValueError('Unable to find XML metadata file for %s' % fname)
    g = GuanoFile(fname)
    with open(xmlfname, 'rt') as f:
        xml = ElementTree.parse(f)

    g['Timestamp'] = get(xml, 'DateTime', lambda x: datetime.strptime(x, '%d.%m.%Y %H:%M:%S'))
    g['Firmware Version'] = get(xml, 'Firmware')
    g['Make'] = 'Elekon'
    g['Model'] = 'BatLogger'
    g['Serial'] = get(xml, 'SN')
    g['Samplerate'] = get(xml, 'Samplerate', lambda x: int(x.split()[0]))
    g['Length'] = get(xml, 'Duration', lambda x: float(x.split()[0]))
    g['Original Filename'] = get(xml, 'Filename')
    g['Temperature Ext'] = get(xml, 'Temperature', lambda x: float(x.split()[0]))
    g['Loc Position'] = get(xml, 'GPS/Position', lambda x: tuple(map(float, x.split())))
    g['Loc Elevation'] = get(xml, 'GPS/Altitude', lambda x: float(x.split()[0]))

    g['Elekon|BattVoltage'] = get(xml, 'BattVoltage')
    for node in xml.find('Trigger'):
        g['Elekon|Trigger|%s' % node.tag] = node.text
    for node in xml.find('GPS'):
        g['Elekon|GPS|%s' % node.tag] = node.text

    # for k, v in g.items():
    #     print('%s:\t%s' % (k, v))

    print(g.to_string())
    g.write()
    os.remove(xmlfname)

    return g


if __name__ == '__main__':
    from glob import glob

    if len(sys.argv) < 2:
        print('usage: %s FILE...' % os.path.basename(sys.argv[0]), file=sys.stderr)
        sys.exit(2)

    if os.name == 'nt' and '*' in sys.argv[1]:
        fnames = glob(sys.argv[1])
    else:
        fnames = sys.argv[1:]

    for fname in fnames:
        print(fname, '...')
        batlogger2guano(fname)
        print()
